from typing import Dict, Optional

import tensorflow as tf


class DataContainer:
    def __init__(
        self,
        base: tf.data.Dataset,
        label_map: Dict[str, int],
        train: Optional[tf.data.Dataset] = None,
        train_steps: Optional[int] = None,
        validation: Optional[tf.data.Dataset] = None,
        validation_steps: Optional[int] = None,
        test: Optional[tf.data.Dataset] = None,
        test_steps: Optional[int] = None,
        n_classes: Optional[int] = None,
    ):
        self.base = base
        self.train = train
        self.train_steps = train_steps
        self.validation = validation
        self.validation_steps = validation_steps
        self.test = test
        self.test_steps = test_steps
        self.label_map = label_map
        if n_classes:
            self.n_classes = n_classes
        elif label_map:
            self.n_classes = len(label_map.keys())



import os
import re
from pathlib import Path
from typing import Optional, Union

import numpy as np
import tensorflow as tf

from ..data import DataBundle


class ImageDataBundle(DataBundle):
    @classmethod
    def from_subfolders(cls, path: Union[Path, str]) -> "ImageDataBundle":
        path = Path(path)
        paths = []
        labels = []
        for label in os.listdir(path):
            for image_path in os.listdir(path / label):
                paths.append(str(path / label / image_path))
                labels.append(label)

        return cls(np.asarray(paths), np.asarray(labels))

    @classmethod
    def from_re(
        cls,
        path: Union[Path, str],
        regex: str,
        default: Optional[Union[int, float, str, bool]] = None,
    ) -> "ImageDataBundle":
        paths = []
        labels = []
        for value in os.listdir(path):
            match = re.match(regex, value)
            if match:
                labels.append(match.group(1))
            elif default:
                labels.append(str(default))
            else:
                raise ValueError(
                    f"No match found and no default value provided for value: {value}"
                )
            paths.append(f"{path}/{value}")

        return cls(np.asarray(paths), np.asarray(labels))



import attr

import tensorflow as tf
from tensorflow import keras


@attr.s(auto_attribs=True)
class ImageTrainingCycle:
    data: tf.data.Dataset
    steps: int
    n_epochs: int
    lr: float
    optimizer: keras.optimizers.Optimizer
    freeze: bool = False



import matplotlib.pyplot as plt

from ..data import DataContainer


class ImageDataContainer(DataContainer):
    def show(self, mode: str, rows: int = 3, cols: int = 6, debug: bool = False):
        if debug and self.label_map:
            reverse_label_map = {value: key for key, value in self.label_map.items()}
        figsize = (4 * cols, 5 * rows) if debug else (3 * cols, 3 * rows)
        _, ax = plt.subplots(rows, cols, figsize=figsize)

        for i, (x, y) in enumerate(getattr(self, mode).unbatch().take(rows * cols)):
            x = x.numpy()
            y = y.numpy()
            idx = (i // cols, i % cols) if rows > 1 else i % cols
            ax[idx].axis("off")
            ax[idx].imshow(x)
            if debug and self.label_map:
                title = (
                    f"{reverse_label_map[y][:40]}\nLabel code: {y}\nShape: {x.shape}"
                )
            elif debug:
                title = f"Label code: {y}\nShape: {x.shape}"
            else:
                title = y
            ax[idx].set_title(title)



import numpy as np
from sklearn.metrics import mean_squared_error


def rmse(y_true, y_pred, sample_weight=None, multioutput="uniform_average"):
    return np.sqrt(
        mean_squared_error(
            y_true=y_true,
            y_pred=y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )
    )



import time
from typing import Iterable, Optional

import attr
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.metrics import classification_report

from .ImageTrainingCycle import ImageTrainingCycle
from .ImageLearner import ImageLearner
from ..data import DataContainer


@attr.s(auto_attribs=True)
class ImageTrainer:
    learner: ImageLearner
    data_container: DataContainer

    def train(
        self,
        cycles: Iterable[ImageTrainingCycle],
        template: str = (
            "Name: {} Train Time: {:.1f} min. "
            "Eval Time: {:.2f}s Loss: {:.4f} Accuracy: {:.2%}"
        ),
    ):
        start_time = time.time()
        for cycle in cycles:
            self.learner.freeze() if cycle.freeze else self.learner.unfreeze()
            self.learner.compile(optimizer=cycle.optimizer, lr=cycle.lr)
            self.learner.fit(
                cycle.n_epochs,
                cycle.data,
                cycle.steps,
                self.data_container.validation,
                self.data_container.validation_steps,
            )
        end_time = time.time()

        eval_start_time = time.time()
        evaluation_results = self.evaluate(
            self.data_container.validation,
            self.data_container.validation_steps,
            verbose=0,
        )
        eval_end_time = time.time()

        print("-".center(80, "-"))
        print(
            template.format(
                self.learner.base_model.name,
                (end_time - start_time) / 60,
                (eval_end_time - eval_start_time),
                *evaluation_results,
            )
        )
        print("-".center(80, "-"))

    def evaluate(
        self, dataset: tf.data.Dataset, steps: Optional[int] = None, verbose: int = 1
    ):
        return self.learner.model.evaluate(dataset, steps=steps, verbose=verbose)

    def predict(
        self, dataset: tf.data.Dataset, steps: Optional[int] = None, verbose: int = 0
    ):
        return self.learner.model.predict(dataset, steps=steps, verbose=verbose)

    def report(
        self, dataset: tf.data.Dataset, steps: Optional[int] = None, verbose: int = 0
    ):
        return classification_report(
            [label.numpy() for _, label in dataset.take(steps).unbatch()],
            self.learner.model.predict(dataset, steps=steps).argmax(axis=1),
        )

    def analyse(
        self, dataset: tf.data.Dataset, steps: Optional[int] = None, verbose: int = 0
    ):
        reverse_label_map = {
            value: key for key, value in self.data_container.label_map.items()
        }
        images = []
        label_codes = []
        for image, label_code in dataset.take(steps).unbatch():
            label_codes.append(label_code.numpy())
            images.append(image.numpy())
        labels = [reverse_label_map[label_code] for label_code in label_codes]
        probs = self.learner.model.predict(dataset, steps=steps)
        pred_codes = probs.argmax(axis=1)
        preds = [reverse_label_map[pred_code] for pred_code in pred_codes]
        return pd.DataFrame.from_dict(
            {
                "image": images,
                "label": labels,
                "label_code": label_codes,
                "pred": preds,
                "pred_code": pred_codes,
                "label_probs": probs[:, label_codes][np.eye(len(labels), dtype=bool)],
                "pred_probs": probs[:, pred_codes][np.eye(len(pred_codes), dtype=bool)],
            }
        )

    def show_predictions(
        self,
        dataset: tf.data.Dataset,
        steps: int,
        correct: bool = False,
        ascending: bool = True,
        rows: int = 4,
        cols: int = 4,
    ):
        df = self.analyse(dataset=dataset, steps=steps)
        df = df[(df.label == df.pred) if correct else (df.label != df.pred)]
        df.sort_values(by=["label_probs"], ascending=ascending, inplace=True)
        _, ax = plt.subplots(rows, cols, figsize=(4 * cols, 5 * rows))
        for i, row in enumerate(df.head(cols * rows).itertuples()):
            idx = (i // cols, i % cols) if rows > 1 else i % cols
            ax[idx].axis("off")
            ax[idx].imshow(row.image)
            ax[idx].set_title(
                f"{row.label}\n{row.pred}\n{row.label_probs:.4f}\n{row.pred_probs:.4f}"
            )



import math
from typing import Dict, Optional, Tuple, Sequence

import numpy as np
import pandas as pd
import sklearn
import tensorflow as tf


class DataBundle:
    def __init__(self, x: np.ndarray, y: np.ndarray):
        self.x = x
        self.y = y

    @classmethod
    def from_unbalanced(
        cls,
        data_bundle: "DataBundle",
        target_class_size: int,
        value_counts: Dict[int, int],
    ) -> "DataBundle":
        x = []
        y = []
        for label, current_size in value_counts.items():
            current_indices = np.argwhere(data_bundle.y == label).flatten()
            next_indices = []
            for _ in range(target_class_size // current_size):
                next_indices.append(np.random.permutation(current_indices))
            next_indices.append(
                np.random.choice(current_indices, target_class_size % current_size)
            )
            next_indices = np.concatenate(next_indices)
            x.append(data_bundle.x[next_indices])
            y.append(data_bundle.y[next_indices])

        return cls(np.concatenate(x), np.concatenate(y))

    @classmethod
    def split(
        cls,
        data_bundle: "DataBundle",
        fracs: Sequence[float],
        random: Optional[bool] = True,
    ) -> Tuple["DataBundle", ...]:
        x = data_bundle.x
        y = data_bundle.y

        if random:
            random_indices = np.random.permutation(len(data_bundle))
            x = x[random_indices]
            y = y[random_indices]

        result = []
        current_index = 0
        for frac in fracs:
            dx = math.ceil(len(data_bundle) * frac)
            split_data_bundle = cls(
                x=x[current_index : current_index + dx],
                y=y[current_index : current_index + dx],
            )
            result.append(split_data_bundle)
            current_index += dx

        return tuple(result)

    @classmethod
    def from_dataframe(
        cls, dataframe: pd.DataFrame, x_col: str, y_col: str
    ) -> "DataBundle":
        return cls(dataframe[x_col].values, dataframe[y_col].values)

    def __len__(self) -> int:
        return len(self.y)

    def to_dataset(self) -> tf.data.Dataset:
        return tf.data.Dataset.from_tensor_slices((self.x, self.y))

    def value_counts(self) -> Dict[str, int]:
        values, counts = np.unique(self.y, return_counts=True)
        return dict(zip(values.tolist(), counts.tolist()))

    def make_label_map(self) -> Dict[str, int]:
        return {value: key for key, value in enumerate(np.unique(self.y))}

    def apply_label_map(self, label_map: Dict[str, int]) -> None:
        self.y = np.vectorize(label_map.get)(self.y)

    def make_label_scaler(self) -> sklearn.base.BaseEstimator:
        scaler = sklearn.preprocessing.RobustScaler()
        scaler.fit(self.y)
        return scaler
